import logging
import math

from bdffont import BdfFont, BdfGlyph, xlfd

import pixel_font_builder
from pixel_font_builder.glyph import Glyph
from pixel_font_builder.info import SerifMode, WidthMode

logger = logging.getLogger('pixel_font_builder.bdf')


class Config:
    def __init__(
            self,
            resolution_x: int = 75,
            resolution_y: int = 75,
    ):
        self.resolution_x = resolution_x
        self.resolution_y = resolution_y


def _create_glyph(
        font_size: int,
        config: Config,
        code_point: int,
        glyph: Glyph,
) -> BdfGlyph:
    scalable_width_x = math.ceil((glyph.advance_width / font_size) * (72 / config.resolution_x) * 1000)
    return BdfGlyph(
        name=glyph.name,
        code_point=code_point,
        scalable_width=(scalable_width_x, 0),
        device_width=(glyph.advance_width, 0),
        bounding_box_size=glyph.dimensions,
        bounding_box_offset=glyph.horizontal_origin,
        bitmap=glyph.data,
    )


def create_font(context: 'pixel_font_builder.FontBuilder') -> BdfFont:
    config = context.bdf_config
    font_size = context.size
    meta_info = context.meta_info
    horizontal_header = context.horizontal_header
    os2_config = context.os2_config
    character_mapping = context.character_mapping
    name_to_glyph = context.prepare_glyphs()[1]

    logger.debug("Create 'BdfFont': %s", meta_info.family_name)
    font = BdfFont(
        point_size=font_size,
        resolution_xy=(config.resolution_x, config.resolution_y),
        bounding_box_size=(font_size, horizontal_header.line_height),
        bounding_box_offset=(0, horizontal_header.descent),
    )

    logger.debug("Add 'Glyph': .notdef")
    font.add_glyph(_create_glyph(font_size, config, -1, name_to_glyph['.notdef']))
    for code_point, glyph_name in character_mapping.items():
        logger.debug("Add 'Glyph': %s", glyph_name)
        font.add_glyph(_create_glyph(font_size, config, code_point, name_to_glyph[glyph_name]))

    logger.debug("Setup 'Properties'")
    font.properties.foundry = meta_info.manufacturer
    font.properties.family_name = meta_info.family_name
    font.properties.weight_name = meta_info.style_name
    font.properties.slant = xlfd.Slant.ROMAN
    font.properties.setwidth_name = xlfd.SetwidthName.NORMAL
    if meta_info.serif_mode == SerifMode.SERIF:
        font.properties.add_style_name = xlfd.AddStyleName.SERIF
    elif meta_info.serif_mode == SerifMode.SANS_SERIF:
        font.properties.add_style_name = xlfd.AddStyleName.SANS_SERIF
    else:
        font.properties.add_style_name = meta_info.serif_mode
    font.properties.pixel_size = font_size
    font.properties.point_size = font_size * 10
    font.properties.resolution_x = config.resolution_x
    font.properties.resolution_y = config.resolution_y
    if meta_info.width_mode == WidthMode.MONOSPACED:
        font.properties.spacing = xlfd.Spacing.MONOSPACED
    elif meta_info.width_mode == WidthMode.DUOSPACED:
        font.properties.spacing = 'D'
    elif meta_info.width_mode == WidthMode.PROPORTIONAL:
        font.properties.spacing = xlfd.Spacing.PROPORTIONAL
    else:
        font.properties.spacing = meta_info.width_mode
    font.properties.average_width = round(sum([glyph.device_width_x * 10 for glyph in font.code_point_to_glyph.values()]) / font.get_glyphs_count())
    font.properties.charset_registry = xlfd.CharsetRegistry.ISO10646
    font.properties.charset_encoding = '1'

    font.properties.default_char = -1
    font.properties.font_ascent = horizontal_header.ascent
    font.properties.font_descent = horizontal_header.descent
    font.properties.x_height = os2_config.x_height
    font.properties.cap_height = os2_config.cap_height

    font.properties.font_version = meta_info.version
    font.properties.copyright = meta_info.copyright_info
    font.properties['LICENSE'] = meta_info.license_info

    font.generate_xlfd_font_name()

    logger.debug("Create 'BdfFont' finished")
    return font



import os
import shutil

from examples import build_dir
from pixel_font_builder import FontBuilder, StyleName, SerifMode, WidthMode, Glyph, opentype


def main():
    outputs_dir = os.path.join(build_dir, 'create')
    if os.path.exists(outputs_dir):
        shutil.rmtree(outputs_dir)
    os.makedirs(outputs_dir)

    builder = FontBuilder(12)

    builder.meta_info.version = '1.0.0'
    builder.meta_info.family_name = 'My Pixel'
    builder.meta_info.style_name = StyleName.REGULAR
    builder.meta_info.serif_mode = SerifMode.SANS_SERIF
    builder.meta_info.width_mode = WidthMode.MONOSPACED
    builder.meta_info.manufacturer = 'Pixel Font Studio'
    builder.meta_info.designer = 'TakWolf'
    builder.meta_info.description = 'A demo pixel font.'
    builder.meta_info.copyright_info = 'Copyright (c) TakWolf'
    builder.meta_info.license_info = 'This Font Software is licensed under the SIL Open Font License, Version 1.1.'
    builder.meta_info.vendor_url = 'https://github.com/TakWolf/pixel-font-builder'
    builder.meta_info.designer_url = 'https://takwolf.com'
    builder.meta_info.license_url = 'https://openfontlicense.org'
    builder.meta_info.sample_text = 'Hello World!'

    builder.horizontal_header.ascent = 10
    builder.horizontal_header.descent = -2

    builder.vertical_header.ascent = 10
    builder.vertical_header.descent = -2

    builder.os2_config.x_height = 5
    builder.os2_config.cap_height = 7

    builder.character_mapping.update({
        ord('A'): 'CAP_LETTER_A',
    })

    builder.glyphs.append(Glyph(
        name='.notdef',
        advance_width=8,
        advance_height=12,
        horizontal_origin=(0, -2),
        vertical_origin_y=0,
        data=[
            [1, 1, 1, 1, 1, 1, 1, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 1],
            [1, 1, 1, 1, 1, 1, 1, 1],
        ],
    ))
    builder.glyphs.append(Glyph(
        name='CAP_LETTER_A',
        advance_width=8,
        advance_height=12,
        horizontal_origin=(0, -2),
        vertical_origin_y=0,
        data=[
            [0, 0, 0, 1, 1, 0, 0, 0],
            [0, 0, 1, 0, 0, 1, 0, 0],
            [0, 0, 1, 0, 0, 1, 0, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 1, 1, 1, 1, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 1, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
        ],
    ))

    builder.save_otf(os.path.join(outputs_dir, 'my-pixel.otf'))
    builder.save_otf(os.path.join(outputs_dir, 'my-pixel.woff2'), flavor=opentype.Flavor.WOFF2)
    builder.save_ttf(os.path.join(outputs_dir, 'my-font.ttf'))
    builder.save_bdf(os.path.join(outputs_dir, 'my-font.bdf'))


if __name__ == '__main__':
    main()




class Glyph:
    def __init__(
            self,
            name: str,
            advance_width: int = 0,
            advance_height: int = 0,
            horizontal_origin: tuple[int, int] = (0, 0),
            vertical_origin_y: int = 0,
            data: list[list[int]] = None,
    ):
        self.name = name
        self.advance_width = advance_width
        self.advance_height = advance_height
        self.horizontal_origin_x, self.horizontal_origin_y = horizontal_origin
        self.vertical_origin_y = vertical_origin_y
        if data is None:
            data = []
        self.data = data

    @property
    def horizontal_origin(self) -> tuple[int, int]:
        return self.horizontal_origin_x, self.horizontal_origin_y

    @horizontal_origin.setter
    def horizontal_origin(self, value: tuple[int, int]):
        self.horizontal_origin_x, self.horizontal_origin_y = value

    @property
    def width(self) -> int:
        if len(self.data) > 0:
            return len(self.data[0])
        else:
            return 0

    @property
    def height(self) -> int:
        return len(self.data)

    @property
    def dimensions(self) -> tuple[int, int]:
        return self.width, self.height

    def calculate_top_side_bearing(self) -> int:
        top_empty = 0
        for data_row in self.data:
            if any(data_row) != 0:
                break
            top_empty += 1
        return top_empty - self.vertical_origin_y

    def check_ready(self):
        if any(len(data_row) != len(self.data[0]) for data_row in self.data):
            raise Exception(f"Glyph '{self.name}': lengths of data rows are not equals")



import datetime
import math
import os
import shutil

import png

from examples import glyphs_dir, build_dir
from pixel_font_builder import FontBuilder, FontCollectionBuilder, StyleName, SerifMode, WidthMode, Glyph, opentype


def _load_glyph_data_from_png(file_path: str) -> tuple[list[list[int]], int, int]:
    width, height, bitmap, _ = png.Reader(filename=file_path).read()
    data = []
    for bitmap_row in bitmap:
        data_row = []
        for x in range(0, width * 4, 4):
            alpha = bitmap_row[x + 3]
            if alpha > 127:
                data_row.append(1)
            else:
                data_row.append(0)
        data.append(data_row)
    return data, width, height


def _save_glyph_data_to_png(data: list[list[int]], file_path: str):
    bitmap = []
    for data_row in data:
        bitmap_row = []
        for x in data_row:
            bitmap_row.append(0)
            bitmap_row.append(0)
            bitmap_row.append(0)
            if x == 0:
                bitmap_row.append(0)
            else:
                bitmap_row.append(255)
        bitmap.append(bitmap_row)
    png.from_array(bitmap, 'RGBA').save(file_path)


class GlyphFile:
    @staticmethod
    def load(file_path: str) -> 'GlyphFile':
        hex_name = os.path.basename(file_path).removesuffix('.png')
        if hex_name == 'notdef':
            code_point = -1
        else:
            code_point = int(hex_name, 16)
        return GlyphFile(file_path, code_point)

    def __init__(self, file_path: str, code_point: int):
        self.file_path = file_path
        self.code_point = code_point
        self.glyph_data, self.glyph_width, self.glyph_height = _load_glyph_data_from_png(file_path)

    @property
    def glyph_name(self) -> str:
        if self.code_point == -1:
            return '.notdef'
        else:
            return f'uni{self.code_point:04X}'

    def save(self):
        _save_glyph_data_to_png(self.glyph_data, self.file_path)


def _collect_glyph_files(root_dir: str) -> tuple[dict[int, str], list[GlyphFile]]:
    registry = {}
    for file_dir, _, file_names in os.walk(root_dir):
        for file_name in file_names:
            if not file_name.endswith('.png'):
                continue
            file_path = os.path.join(file_dir, file_name)
            glyph_file = GlyphFile.load(file_path)
            registry[glyph_file.code_point] = glyph_file

    character_mapping = {}
    glyph_files = []
    for glyph_file in registry.values():
        if glyph_file.code_point != -1:
            character_mapping[glyph_file.code_point] = glyph_file.glyph_name
        glyph_files.append(glyph_file)
    glyph_files.sort(key=lambda x: x.code_point)

    for code_point in range(ord('A'), ord('Z') + 1):
        fallback_code_point = code_point + ord('Ａ') - ord('A')
        character_mapping[fallback_code_point] = character_mapping[code_point]
    for code_point in range(ord('a'), ord('z') + 1):
        fallback_code_point = code_point + ord('ａ') - ord('a')
        character_mapping[fallback_code_point] = character_mapping[code_point]
    for code_point in range(ord('0'), ord('9') + 1):
        fallback_code_point = code_point + ord('０') - ord('0')
        character_mapping[fallback_code_point] = character_mapping[code_point]

    return character_mapping, glyph_files


def _create_builder(
        glyph_pool: dict[str, Glyph],
        character_mapping: dict[int, str],
        glyph_files: list[GlyphFile],
        name_num: int = None,
) -> FontBuilder:
    builder = FontBuilder(11)

    time = datetime.datetime.fromisoformat('2024-01-01T00:00:00Z')
    builder.created_time = time
    builder.modified_time = time

    builder.meta_info.version = '1.0.0'
    builder.meta_info.family_name = 'Demo Pixel'
    if name_num is not None:
        builder.opentype_config.cff_family_name = builder.meta_info.family_name
        builder.meta_info.family_name += f' {name_num}'
    builder.meta_info.style_name = StyleName.REGULAR
    builder.meta_info.serif_mode = SerifMode.SANS_SERIF
    builder.meta_info.width_mode = WidthMode.PROPORTIONAL
    builder.meta_info.manufacturer = 'Pixel Font Studio'
    builder.meta_info.designer = 'TakWolf'
    builder.meta_info.description = 'A demo pixel font.'
    builder.meta_info.copyright_info = 'Copyright (c) TakWolf'
    builder.meta_info.license_info = 'This Font Software is licensed under the SIL Open Font License, Version 1.1.'
    builder.meta_info.vendor_url = 'https://github.com/TakWolf/pixel-font-builder'
    builder.meta_info.designer_url = 'https://takwolf.com'
    builder.meta_info.license_url = 'https://openfontlicense.org'
    builder.meta_info.sample_text = 'Hello World!'

    builder.horizontal_header.ascent = 11
    builder.horizontal_header.descent = -4

    builder.vertical_header.ascent = 11
    builder.vertical_header.descent = -4

    builder.os2_config.x_height = 5
    builder.os2_config.cap_height = 7

    builder.character_mapping.update(character_mapping)

    for glyph_file in glyph_files:
        if glyph_file.file_path in glyph_pool:
            glyph = glyph_pool[glyph_file.file_path]
        else:
            horizontal_origin_y = math.floor((builder.horizontal_header.ascent + builder.horizontal_header.descent - glyph_file.glyph_height) / 2)
            vertical_origin_y = (glyph_file.glyph_height - builder.size) // 2
            glyph = Glyph(
                name=glyph_file.glyph_name,
                advance_width=glyph_file.glyph_width,
                advance_height=builder.size,
                horizontal_origin=(0, horizontal_origin_y),
                vertical_origin_y=vertical_origin_y,
                data=glyph_file.glyph_data,
            )
            glyph_pool[glyph_file.file_path] = glyph
        builder.glyphs.append(glyph)

    return builder


def main():
    outputs_dir = os.path.join(build_dir, 'demo')
    if os.path.exists(outputs_dir):
        shutil.rmtree(outputs_dir)
    os.makedirs(outputs_dir)

    character_mapping, glyph_files = _collect_glyph_files(glyphs_dir)
    for glyph_file in glyph_files:
        glyph_file.save()
    glyph_pool = {}

    builder = _create_builder(glyph_pool, character_mapping, glyph_files)
    builder.save_otf(os.path.join(outputs_dir, 'demo.otf'))
    builder.save_otf(os.path.join(outputs_dir, 'demo.woff2'), flavor=opentype.Flavor.WOFF2)
    builder.save_ttf(os.path.join(outputs_dir, 'demo.ttf'))
    builder.save_bdf(os.path.join(outputs_dir, 'demo.bdf'))

    collection_builder = FontCollectionBuilder()
    for index in range(100):
        builder = _create_builder(glyph_pool, character_mapping, glyph_files, index)
        collection_builder.font_builders.append(builder)
    collection_builder.save_otc(os.path.join(outputs_dir, 'demo.otc'))
    collection_builder.save_ttc(os.path.join(outputs_dir, 'demo.ttc'))


if __name__ == '__main__':
    main()



from pixel_font_builder.font import FontBuilder, FontCollectionBuilder
from pixel_font_builder.info import StyleName, SerifMode, WidthMode, MetaInfo, LayoutHeader
from pixel_font_builder.glyph import Glyph



import datetime
import os

import bdffont
import fontTools.fontBuilder
import fontTools.ttLib

from pixel_font_builder import os2, opentype, bdf
from pixel_font_builder.glyph import Glyph
from pixel_font_builder.info import MetaInfo, LayoutHeader


class FontBuilder:
    def __init__(self, size: int):
        self.size = size
        self.created_time: datetime.datetime | None = None
        self.modified_time: datetime.datetime | None = None

        self.meta_info = MetaInfo()
        self.horizontal_header = LayoutHeader()
        self.vertical_header = LayoutHeader()
        self.os2_config = os2.Config()
        self.character_mapping: dict[int, str] = {}
        self.glyphs: list[Glyph] = []
        self.opentype_config = opentype.Config()
        self.bdf_config = bdf.Config()

    def prepare_glyphs(self) -> tuple[list[str], dict[str, Glyph]]:
        glyph_order = ['.notdef']
        name_to_glyph = {}

        for glyph in self.glyphs:
            if glyph.name in name_to_glyph:
                raise Exception(f"Duplicate glyphs: '{glyph.name}'")
            glyph.check_ready()
            if glyph.name != '.notdef':
                glyph_order.append(glyph.name)
            name_to_glyph[glyph.name] = glyph

        if '.notdef' not in name_to_glyph:
            raise Exception("Need to provide a glyph named '.notdef'")

        for code_point, glyph_name in self.character_mapping.items():
            if code_point < 0:
                raise Exception('Code points must >= 0')
            if glyph_name not in name_to_glyph:
                raise Exception(f"Missing glyph: '{glyph_name}'")

        return glyph_order, name_to_glyph

    def to_xtf_builder(self, is_ttf: bool, flavor: opentype.Flavor = None) -> fontTools.fontBuilder.FontBuilder:
        return opentype.create_builder(self, is_ttf, flavor)

    def to_otf_builder(self, flavor: opentype.Flavor = None) -> fontTools.fontBuilder.FontBuilder:
        return self.to_xtf_builder(False, flavor)

    def save_otf(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            flavor: opentype.Flavor = None,
    ):
        self.to_otf_builder(flavor).save(file_path)

    def to_ttf_builder(self, flavor: opentype.Flavor = None) -> fontTools.fontBuilder.FontBuilder:
        return self.to_xtf_builder(True, flavor)

    def save_ttf(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            flavor: opentype.Flavor = None,
    ):
        self.to_ttf_builder(flavor).save(file_path)

    def to_bdf_builder(self) -> bdffont.BdfFont:
        return bdf.create_font(self)

    def save_bdf(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            optimize_bitmap: bool = True,
    ):
        self.to_bdf_builder().save(file_path, optimize_bitmap)


class FontCollectionBuilder:
    def __init__(self, font_builders: list[FontBuilder] = None):
        if font_builders is None:
            font_builders = []
        self.font_builders = font_builders

    def _to_xtf_builders(self, is_ttf: bool) -> list[fontTools.fontBuilder.FontBuilder]:
        return [font_builder.to_xtf_builder(is_ttf) for font_builder in self.font_builders]

    def to_xtc_builder(self, is_ttf: bool) -> fontTools.ttLib.TTCollection:
        return opentype.create_collection_builder(self._to_xtf_builders(is_ttf))

    def to_otc_builder(self) -> fontTools.ttLib.TTCollection:
        return self.to_xtc_builder(False)

    def save_otc(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            share_tables: bool = True,
    ):
        self.to_otc_builder().save(file_path, share_tables)

    def to_ttc_builder(self) -> fontTools.ttLib.TTCollection:
        return self.to_xtc_builder(True)

    def save_ttc(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            share_tables: bool = True,
    ):
        self.to_ttc_builder().save(file_path, share_tables)




class Config:
    def __init__(
            self,
            x_height: int = 0,
            cap_height: int = 0,
    ):
        self.x_height = x_height
        self.cap_height = cap_height

    def __mul__(self, other: int) -> 'Config':
        return Config(
            self.x_height * other,
            self.cap_height * other,
        )



import logging
import os

logging.basicConfig(level=logging.DEBUG)

project_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
glyphs_dir = os.path.join(project_root_dir, 'assets', 'glyphs')
build_dir = os.path.join(project_root_dir, 'build')
